package com.fzm.common.service.impl;

import cn.hutool.core.util.IdUtil;
import cn.hutool.json.JSONObject;
import cn.hutool.json.JSONUtil;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ObjectNode;
import com.fzm.common.constant.SystemConstant;
import com.fzm.common.entity.Order;
import com.fzm.common.entity.Payment;
import com.fzm.common.entity.Refund;
import com.fzm.common.entity.dto.JsapiPayDto;
import com.fzm.common.entity.dto.OrderProcessMsg;
import com.fzm.common.enums.*;
import com.fzm.common.exception.GlobalException;
import com.fzm.common.properties.WxPayProperties;
import com.fzm.common.service.*;
import com.fzm.common.utils.SnowflakeUtil;
import com.wechat.pay.contrib.apache.httpclient.util.AesUtil;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.apache.http.client.methods.CloseableHttpResponse;
import org.apache.http.client.methods.HttpGet;
import org.apache.http.client.methods.HttpPost;
import org.apache.http.entity.StringEntity;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.http.util.EntityUtils;
import org.redisson.Redisson;
import org.redisson.api.RLock;
import org.springframework.amqp.rabbit.core.RabbitTemplate;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.Base64Utils;

import javax.annotation.Resource;
import javax.servlet.http.HttpServletRequest;
import java.io.BufferedReader;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.security.GeneralSecurityException;
import java.security.PrivateKey;
import java.security.Signature;
import java.time.Instant;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.TimeUnit;

/**
 * @author tangtuo
 * @date 2022/1/13 10:13
 */
@Slf4j
@Service
public class WxPayServiceImpl implements WxPayService {
    @Resource
    private CloseableHttpClient httpClient;
    @Resource
    private WxPayProperties wxPayProperties;
    @Resource
    private PrivateKey wxPrivateKey;
    @Resource
    private OrderService orderService;
    @Resource
    private PaymentService paymentService;
    @Resource
    private RabbitTemplate rabbitTemplate;
    @Resource
    private SnowflakeUtil snowflakeUtil;
    @Resource
    private SmsService smsService;
    @Resource
    private Redisson redisson;
    @Resource
    private RefundService refundService;

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Map<String, Object> payJsapi(JsapiPayDto jsapiPayDto) {
        RLock lock = redisson.getLock("pay-" + jsapiPayDto.getOrderId());
        boolean tryLock;
        try {
            tryLock = lock.tryLock(10, TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            throw GlobalException.newException(ResultCode.PAY_FAILED, "当前订单正在支付中,请勿重复点击");
        }
        if (!tryLock) {
            throw GlobalException.newException(ResultCode.PAY_FAILED, "当前订单正在支付中,请勿重复点击");
        }
        try {
            String appId = wxPayProperties.getAppId();
            Long out_trade_no = jsapiPayDto.getOrderId();
            Order order = orderService.getById(out_trade_no);
            // 请求jsapi支付下单接口
            String url = "https://api.mch.weixin.qq.com/v3/pay/transactions/jsapi";
            HttpPost httpPost = new HttpPost(url);
            httpPost.addHeader("Accept", "application/json");
            httpPost.addHeader("Content-type", "application/json; charset=utf-8");
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            ObjectMapper objectMapper = new ObjectMapper();
            ObjectNode rootNode = objectMapper.createObjectNode();
            rootNode.put("mchid", wxPayProperties.getMchId())
                    .put("appid", appId)
                    .put("description", PayScene.getTypeByCode(order.getPayScene()))
                    .put("notify_url", wxPayProperties.getPayNotifyUrl())
                    .put("out_trade_no", String.valueOf(out_trade_no));
            rootNode.putObject("amount")
                    .put("total", order.getFee());
            rootNode.putObject("payer")
                    .put("openid", jsapiPayDto.getOpenid());

            JSONObject jsonObject;
            try {
                String json = objectMapper.writeValueAsString(rootNode);
                log.info("h5下单接口请求参数: {}", json);
                objectMapper.writeValue(bos, rootNode);
                httpPost.setEntity(new StringEntity(bos.toString("UTF-8"), "UTF-8"));
                CloseableHttpResponse response = httpClient.execute(httpPost);
                String bodyAsString = EntityUtils.toString(response.getEntity());
                jsonObject = JSONUtil.parseObj(bodyAsString);
            } catch (Exception e) {
                throw GlobalException.newException(ResultCode.PAY_FAILED, "微信支付下单失败");
            }

            if (StringUtils.isBlank(jsonObject.getStr("prepay_id"))) {
                throw GlobalException.newException(ResultCode.PAY_FAILED, "微信支付下单失败");
            }
            String prepay_id = jsonObject.getStr("prepay_id");

            String timeStamp = String.valueOf(Instant.now().getEpochSecond());
            String nonceStr = IdUtil.simpleUUID();
            String packageInfo = "prepay_id=" + prepay_id;
            String signStr = appId + "\n" +
                    timeStamp + "\n" +
                    nonceStr + "\n" +
                    packageInfo + "\n";
            log.info("签名字符串为: {}", signStr);
            String paySign = encryptByPrivateKey(signStr);
            Map<String, Object> result = new HashMap<>();
            result.put("appId", appId);
            result.put("timeStamp", timeStamp);
            result.put("nonceStr", nonceStr);
            result.put("package", packageInfo);
            result.put("signType", "RSA");
            result.put("paySign", paySign);
            result.put("orderId", out_trade_no);
            // 修改此订单的支付方式为微信支付
            orderService.updatePayType(jsapiPayDto.getOrderId(), PayType.WXPAY.getCode());
            return result;
        } finally {
            lock.unlock();
        }
    }


    /**
     * 私钥签名
     *
     * @param data 需要加密的数据
     * @return 加密后的数据
     */
    public String encryptByPrivateKey(String data) {
        try {
            Signature signature = Signature.getInstance("SHA256WithRSA");
            signature.initSign(wxPrivateKey);
            signature.update(data.getBytes(StandardCharsets.UTF_8));
            byte[] signed = signature.sign();
            return Base64Utils.encodeToString(signed);
        } catch (Exception e) {
            throw GlobalException.newException(ResultCode.FAILED);
        }
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Boolean notifyH5(HttpServletRequest request) {
        try {
            String requestBodyData = this.getRequestBodyData(request);
            log.info("接收微信支付回调请求, 请求参数: {}", requestBodyData);
            JSONObject jsonObject = JSONUtil.parseObj(requestBodyData);
            JSONObject resource = jsonObject.getJSONObject("resource");
            String plainText = this.decrypt(resource);
            log.info("解密后的明文信息为: {}", plainText);
            JSONObject obj = JSONUtil.parseObj(plainText);
            // 处理订单
            processOrder(obj);
        } catch (Exception e) {
            log.error("微信支付回调异常", e);
            return false;
        }
        return true;
    }


    @Override
    @Transactional(rollbackFor = Exception.class)
    public void processOrder(JSONObject jsonObject) {
        // 获取订单号
        Long out_trade_no = jsonObject.getLong("out_trade_no");
        // 加锁,避免接口幂等性问题
        RLock lock = redisson.getLock("pay-notify-" + out_trade_no);
        // 修改订单状态
        try {
            boolean tryLock;
            try {
                tryLock = lock.tryLock(10, TimeUnit.SECONDS);
            } catch (InterruptedException e) {
                throw GlobalException.newException(ResultCode.FAILED, "操作频繁");
            }
            if (!tryLock) {
                throw GlobalException.newException(ResultCode.FAILED, "操作频繁");
            }

            Order order = orderService.getById(out_trade_no);
            // 如果订单状态不是待支付,直接返回
            if (!OrderStatus.PAYING.getStatus().equals(order.getOrderStatus())) {
                log.warn("当前订单已处理完成, 订单号:==> {}", out_trade_no);
                return;
            }

            // 修改订单状态
            orderService.updateOrderStatus(out_trade_no, OrderStatus.PAYED);
            // 插入支付流水记录
            Payment payment = new Payment();
            payment.setOrderId(out_trade_no);
            payment.setTradeType(jsonObject.getStr("trade_type"));
            payment.setTotalFee(jsonObject.getJSONObject("amount").getInt("payer_total"));
            payment.setTransactionId(jsonObject.getStr("transaction_id"));
            payment.setSuccessTime(jsonObject.getStr("success_time"));
            payment.setContent(JSONUtil.toJsonStr(jsonObject));
            paymentService.save(payment);

            // 订单支付成功后,给mq发送一条消息,处理nft发行或版权申请的状态
            Integer productId = order.getProductId();
            OrderProcessMsg orderProcessMsg = new OrderProcessMsg(out_trade_no, productId, order.getPayScene());
            rabbitTemplate.convertAndSend("order-exchange", "order.process", orderProcessMsg);

        } finally {
            lock.unlock();
        }
    }


    @Override
    @Transactional(rollbackFor = Exception.class)
    public String queryOrder(Long orderId) {
        String url = String.format("https://api.mch.weixin.qq.com/v3/pay/transactions/out-trade-no/%s?mchid=%s", orderId, wxPayProperties.getMchId());
        HttpGet httpGet = new HttpGet(url);
        httpGet.setHeader("Accept", "application/json");
        //完成签名并执行请求
        try {
            CloseableHttpResponse response = httpClient.execute(httpGet);
            String result = EntityUtils.toString(response.getEntity());
            log.info("查询微信支付订单状态, 当前订单号: {}, 返回结果:{}", orderId, result);
            JSONObject jsonObject = JSONUtil.parseObj(result);
            // 支付成功
            if (WxPayStatus.SUCCESS.getStatus().equals(jsonObject.getStr("trade_state"))) {
                processOrder(jsonObject);
            }
            return jsonObject.getStr("trade_state");
        } catch (IOException e) {
            throw GlobalException.newException(ResultCode.SELECT_FAILED);
        }
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public Boolean refund(Long orderId, Integer channel) {
        boolean result;
        RLock lock = redisson.getLock("refund-" + orderId);
        boolean tryLock;
        try {
            tryLock = lock.tryLock(10, TimeUnit.SECONDS);
        } catch (InterruptedException e) {
            throw GlobalException.newException(ResultCode.REFUND_FAILED, "当前订单正在退款中,请勿重复点击");
        }
        if (!tryLock) {
            throw GlobalException.newException(ResultCode.REFUND_FAILED, "当前订单正在退款中,请勿重复点击");
        }
        try {
            Order order = orderService.getById(orderId);
            if (order == null || !order.getOrderStatus().equals(OrderStatus.PAYED.getStatus())) {
                throw GlobalException.newException(ResultCode.REFUND_FAILED, "当前订单未支付成功：" + orderId);
            }
            // 订单金额=0的时候,直接返回
            if (order.getFee() <= 0) {
                // 修改订单状态为退款成功
                orderService.updateOrderStatus(orderId, OrderStatus.REFUNDED);
                return true;
            }
            Refund refund = refundService.getByOrderId(orderId);
            Long out_refund_no;
            if (refund == null) {
                refund = new Refund();
                out_refund_no = snowflakeUtil.snowflakeId();
            } else {
                Integer refundStatus = refund.getRefundStatus();
                if (refundStatus.equals(RefundStatus.REFUNDING.getStatus())) {
                    throw GlobalException.newException(ResultCode.REFUND_FAILED, "当前订单正在退款中,请勿重复点击");
                }
                if (refundStatus.equals(RefundStatus.SUCCESS.getStatus())) {
                    throw GlobalException.newException(ResultCode.REFUND_FAILED, "当前订单已退款完成,请勿重复点击");
                }
                out_refund_no = refund.getId();
            }
            String url = "https://api.mch.weixin.qq.com/v3/refund/domestic/refunds";
            HttpPost httpPost = new HttpPost(url);
            httpPost.addHeader("Accept", "application/json");
            httpPost.addHeader("Content-type", "application/json; charset=utf-8");
            ByteArrayOutputStream bos = new ByteArrayOutputStream();
            ObjectMapper objectMapper = new ObjectMapper();
            ObjectNode rootNode = objectMapper.createObjectNode();
            rootNode.put("out_refund_no", String.valueOf(out_refund_no))
                    .put("notify_url", wxPayProperties.getRefundNotifyUrl())
                    .put("out_trade_no", String.valueOf(orderId));
            rootNode.putObject("amount")
                    .put("refund", order.getFee())
                    .put("total", order.getFee())
                    .put("currency", "CNY");

            JSONObject jsonObject;
            int statusCode;
            try {
                String json = objectMapper.writeValueAsString(rootNode);
                log.info("退款接口请求参数: {}", json);
                objectMapper.writeValue(bos, rootNode);
                httpPost.setEntity(new StringEntity(bos.toString("UTF-8"), "UTF-8"));
                CloseableHttpResponse response = httpClient.execute(httpPost);
                String bodyAsString = EntityUtils.toString(response.getEntity());
                log.info("退款接口返回参数: {}", bodyAsString);
                jsonObject = JSONUtil.parseObj(bodyAsString);
                statusCode = response.getStatusLine().getStatusCode();
            } catch (Exception e) {
                throw GlobalException.newException(ResultCode.REFUND_FAILED);
            }

            // 发起退款成功
            if (statusCode == 200 || statusCode == 204) {
                refund.setRefundId(jsonObject.getStr("refund_id"));
                refund.setTransactionId(jsonObject.getStr("transaction_id"));
                refund.setUserReceivedAccount(jsonObject.getStr("user_received_account"));
                refund.setRefundStatus(RefundStatus.REFUNDING.getStatus());
                // 修改订单状态为退款中
                orderService.updateOrderStatus(orderId, OrderStatus.REFUNDING);
                result = true;
            } else {
                // 发起退款失败
                refund.setRefundStatus(RefundStatus.FAINED.getStatus());
                refund.setFailMessage(jsonObject.getStr("message"));
                result = false;
            }

            // 插入退款信息
            refund.setId(out_refund_no);
            refund.setFee(order.getFee());
            refund.setOrderId(orderId);
            refund.setChannel(channel);
            refundService.saveOrUpdate(refund);
            // 如果是用户端发起的退款, 失败时则需抛出异常,告知系统管理员退款失败
            if (RefundLaunchChannel.USER.getCode().equals(channel) && result) {
                // 异步发送退款短信通知
                smsService.sendRefundSms(orderId);
            }
            return result;
        } finally {
            lock.unlock();
        }
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void notifyRefund(HttpServletRequest request) {
        try {
            String requestBodyData = this.getRequestBodyData(request);
            log.info("接收微信退款接口回调请求, 请求参数: {}", requestBodyData);
            JSONObject jsonObject = JSONUtil.parseObj(requestBodyData);
            JSONObject resource = jsonObject.getJSONObject("resource");
            String plainText = this.decrypt(resource);
            log.info("解密后的明文信息为: {}", plainText);
            JSONObject obj = JSONUtil.parseObj(plainText);
            processRefund(obj);
        } catch (Exception e) {
            throw GlobalException.newException(ResultCode.FAILED);
        }
    }

    @Override
    @Transactional(rollbackFor = Exception.class)
    public void processRefund(JSONObject obj) {
        Long out_trade_no = obj.getLong("out_trade_no");
        Long out_refund_no = obj.getLong("out_refund_no");
        RLock lock = redisson.getLock("refund-" + out_trade_no);
        try {
            boolean tryLock;
            try {
                tryLock = lock.tryLock(10, TimeUnit.SECONDS);
            } catch (InterruptedException e) {
                throw GlobalException.newException(ResultCode.FAILED, "操作频繁");
            }
            if (!tryLock) {
                throw GlobalException.newException(ResultCode.FAILED, "操作频繁");
            }

            Refund refund = refundService.getById(out_refund_no);
            if (RefundStatus.SUCCESS.getStatus().equals(refund.getRefundStatus())) {
                return;
            }
            // 更新退款信息
            refund.setRefundStatus(RefundStatus.SUCCESS.getStatus());
            refund.setSuccessTime(obj.getStr("success_time"));
            refund.setContent(JSONUtil.toJsonStr(obj));
            refund.setFailMessage("");
            refundService.updateById(refund);
            // 更新流水的退款状态
            Payment payment = paymentService.getByOrderId(out_trade_no);
            payment.setIsRefund(SystemConstant.BOOLEAN_DATA_TRUE);
            paymentService.updateById(payment);
            // 修改订单状态为退款成功
            orderService.updateOrderStatus(out_trade_no, OrderStatus.REFUNDED);

        } finally {
            lock.unlock();
        }
    }


    /**
     * 解密
     */
    private String decrypt(JSONObject resource) throws GeneralSecurityException {
        AesUtil aesUtil = new AesUtil(wxPayProperties.getApiV3Key().getBytes(StandardCharsets.UTF_8));
        String ciphertext = resource.getStr("ciphertext");
        String associated_data = resource.getStr("associated_data");
        String nonce = resource.getStr("nonce");
        // 对密文解密
        return aesUtil.decryptToString(associated_data.getBytes(StandardCharsets.UTF_8),
                nonce.getBytes(StandardCharsets.UTF_8),
                ciphertext);
    }


    /**
     * 获取request中body数据
     */
    private String getRequestBodyData(HttpServletRequest request) throws IOException {
        BufferedReader bufferReader = new BufferedReader(request.getReader());
        StringBuilder sb = new StringBuilder();
        String line;
        while ((line = bufferReader.readLine()) != null) {
            sb.append(line);
        }
        bufferReader.close();
        return sb.toString();
    }

}